# Copyright 2020 Resilient Solutions Inc. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import aioredis
import asyncio
import os
from aioredis.pubsub import Receiver
from typing import List

REDIS_HOST = os.environ.get('REDIS_HOST', 'localhost')


class Listener:
    """Object that listens and responds to specific events triggered on the system.

    This class is intended to be used as a base class for creating listeners. To use
    simply inherit and override the ``handle_message`` method.

    Args:
        channels (str, :obj:`list` of :obj:`str`): Event channel(s) to listen on.
            These can be explicit names (e.g., channel:1) or patterns with a
            wildcard (e.g., channel:*) which will listen to all channels that start
            with "channel:".
    """
    def __init__(self, channels: List[str]) -> None:
        loop = asyncio.get_event_loop()
        db = loop.run_until_complete(aioredis.create_redis(f'redis://{REDIS_HOST}'))
        pubsub = Receiver(loop=loop)

        if isinstance(channels, str):
            channels = [channels]
        subs = [pubsub.channel(channel)
                for channel in channels if not channel.endswith('*')]
        pattern_subs = [pubsub.pattern(channel)
                        for channel in channels if channel.endswith('*')]
        if len(subs) > 0:
            loop.run_until_complete(db.subscribe(*subs))
        if len(pattern_subs) > 0:
            loop.run_until_complete(db.psubscribe(*pattern_subs))

        self.pubsub = pubsub
        self.subs = subs
        self.psubs = pattern_subs

    async def handle_message(self, channel, message):
        """Handler for received messages

        Args:
            channel (str): Name of the channel on which the message was received
            message (str): The message received on the channel
        """
        raise NotImplementedError('You must override handle_message to use this class')

    async def run(self) -> None:
        """Listen for messages on subscribed channels"""
        async for channel, message in self.pubsub.iter():
            if channel.is_pattern:
                channel = message[0]
                message = message[1]
            else:
                channel = channel.name
            asyncio.create_task(
                self.handle_message(channel.decode(), message.decode())
            )

    def start(self) -> None:
        print('Starting listener')
        try:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self.run())
        except KeyboardInterrupt:
            print('Stopping Listener')
        finally:
            # Unsubscribe from channels
            db = loop.run_until_complete(aioredis.create_redis(f'redis://{REDIS_HOST}'))
            map(lambda channel: loop.run_until_complete(db.unsubscribe(channel=channel)), self.subs)
            map(lambda pattern: loop.run_until_complete(db.punsubscribe(pattern=pattern)), self.psubs)
            self.pubsub.stop()
            loop.close()



import pytest
from aiomothr import AsyncJobRequest
from asynctest import CoroutineMock, MagicMock, patch


# For mocking async loops
class AsyncIterator:
    def __init__(self, seq):
        self.iter = iter(seq)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self.iter)
        except StopIteration:
            raise StopAsyncIteration


class TestJob:
    def setup_method(self, _):
        self.submit_response = {'submitJob': {'job': {'jobId': 'test', 'status': 'submitted'}}}
        self.query_response = [
            {'job': {'status': 'submitted'}}, 
            {'job': {'status': 'running'}}, 
            {'job': {'status': 'complete'}}, 
            {'job': {'status': 'complete'}}
        ]

    @pytest.mark.asyncio
    @patch('gql.dsl.DSLSchema.mutate')
    @patch('gql.dsl.DSLSchema.query')
    async def test_run_job(self, mock_query, mock_mutate):
        mock_mutate.return_value = self.submit_response
        mock_query.side_effect = self.query_response
        request = AsyncJobRequest(service='test')
        result = await request.run_job()
        assert(result)

    @pytest.mark.asyncio
    @patch('gql.dsl.DSLSchema.mutate')
    @patch('gql.dsl.DSLSchema.query')
    async def test_run_job_fail(self, mock_query, mock_mutate):
        for i in [-1, -2]:
            self.query_response[i]['job']['status'] = 'failed'
            self.query_response[i]['job']['error'] = 'failed'
        mock_mutate.return_value = self.submit_response
        mock_query.side_effect=self.query_response
        request = AsyncJobRequest(service='test')
        with pytest.raises(RuntimeError):
            result = await request.run_job()

    @pytest.mark.asyncio
    @patch('gql.dsl.DSLSchema.mutate')
    @patch('gql.dsl.DSLSchema.query')
    async def test_run_job_fail_return_failed(self, mock_query, mock_mutate):
        for i in [-1, -2]:
            self.query_response[i]['job']['status'] = 'failed'
            self.query_response[i]['job']['error'] = 'failed'
        mock_mutate.return_value = self.submit_response
        mock_query.side_effect = self.query_response
        request = AsyncJobRequest(service='test')
        result = await request.run_job(return_failed=True)
        assert(result['error'] == 'failed')

    @pytest.mark.asyncio
    @patch('aiomothr.request.Client')
    async def test_subscribe(self, mock_client):
        mock_client.return_value.__aenter__.return_value.subscribe.return_value = AsyncIterator(['test'])
        request = AsyncJobRequest(service='test')
        result = await request.subscribe()
        assert(result == 'test')
        
    @pytest.mark.asyncio
    @patch('aiomothr.request.Client')
    async def test_subscribe_messages(self, mock_client):
        mock_client.return_value.__aenter__.return_value.subscribe.return_value = AsyncIterator([f'message {i+1}' for i in range(10)])
        request = AsyncJobRequest(service='test')
        messages = [m async for m in request.subscribe_messages()]
        assert(len(messages) == 10)
        assert(messages[0] == 'message 1')

    @pytest.mark.asyncio
    @patch('gql.dsl.DSLSchema.mutate')
    async def test_login(self, mock_mutate):
        mock_mutate.return_value = {'login':{'token':'access-token','refresh':'refresh-token'}}
        request = AsyncJobRequest(service='test')
        access, refresh = await request.login(username='test', password='password')
        assert(access == 'access-token')
        assert(refresh == 'refresh-token')

    @pytest.mark.asyncio
    @patch('gql.dsl.DSLSchema.mutate')
    async def test_refresh(self, mock_mutate):
        mock_mutate.return_value = {'refresh':{'token':'access-token'}}
        request = AsyncJobRequest(service='test')
        access = await request.refresh_token()
        assert(access == 'access-token')





import os
import pytest
import sys
from asynctest import CoroutineMock, MagicMock, patch
from collections import namedtuple

sys.path.append('.')

from aiomothr import Listener

class MockSender:
    def __init__(self, channel, message, is_pattern=False):
        channel_obj = namedtuple('channel_obj', 'is_pattern name')
        self.channel = channel_obj(is_pattern=is_pattern, name=channel)
        if is_pattern:
            self.message = (channel, message)
        else:
            self.message = message


class TestListener:
    @patch('aioredis.create_redis')
    @patch('aiomothr.listener.Receiver')
    def setup_method(self, _, mock_pubsub, mock_redis):
        messages = [
            MockSender(channel=b'test', message=b'test message'),
            MockSender(channel=b'test:testing', message=b'test message', is_pattern=True),
            MockSender(channel=b'test', message=b'test message2'),
            MockSender(channel=b'test', message=b'test message3')
        ]
        messages = [(msg.channel, msg.message) for msg in messages]
        mock_redis.return_value.subscribe = CoroutineMock()
        mock_redis.return_value.psubscribe = CoroutineMock()
        mock_pubsub.return_value.iter = MagicMock()
        mock_pubsub.return_value.iter.return_value.__aiter__.return_value = [msg for msg in messages]
        self.listener = Listener(['test', 'test:*'])
        self.listener.handle_message = CoroutineMock()

    @patch('aioredis.create_redis')
    def test_run(self, mock_redis):
        self.listener.start()
        assert(self.listener.handle_message.call_count == 4)



import setuptools

setuptools.setup(
    name='aiomothr',
    version='0.1.2',
    author='James Arnold',
    author_email='james@rs21.io',
    description='Asynchronous client library for interacting with MOTHR',
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'aioredis',
        'gql==3.0.0a1',
        'mothrpy>=0.1.0',
    ],
    extras_require={
        'dev': [
            'asynctest',
            'pytest', 
            'pytest-asyncio',
            'pytest-cov',
            'pytest-mypy'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.6'
)
